from abc import ABC, abstractmethod
from typing import Optional

from naneos.partector.blueprints._data_structure import NaneosDeviceDataPoint


class PartectorBleDecoderBlueprint(ABC):
    @classmethod
    @abstractmethod
    def decode(
        cls, data: bytes, data_structure: Optional[NaneosDeviceDataPoint] = None
    ) -> NaneosDeviceDataPoint:
        """
        Decode the advertisement data from the Partector device. If the optional data_structure is
        given, it will be filled with the decoded data and returned.
        """
        pass



import base64
import datetime
import pickle
from threading import Thread
from typing import Callable, ClassVar, Optional

import pandas as pd
import requests

from naneos.logger import LEVEL_WARNING, get_naneos_logger
from naneos.protobuf.protobuf import create_combined_entry, create_proto_device

logger = get_naneos_logger(__name__, LEVEL_WARNING)


class NaneosUploadThread(Thread):
    URL: ClassVar[str] = "https://hg3zkburji.execute-api.eu-central-1.amazonaws.com/prod/proto/v1"
    HEADERS: ClassVar[dict] = {
        "Content-Type": "application/json",
        "Accept": "application/json",
    }

    def __init__(
        self,
        data: dict[int, pd.DataFrame],
        callback: Optional[Callable[[bool], None]],
    ) -> None:
        """Adding the data that should be uploaded to the database.

        Args:
            data (dict[int, pd.DataFrame]): Data to upload, where the key is the device serial number and the value is a DataFrame.
            callback (Optional[Callable[[bool], None]]): Callback function that is called after upload.
        """
        super().__init__()
        self.data = data
        self._callback = callback

    def run(self) -> None:
        try:
            ret = self.upload(self.data)

            if self._callback:
                if ret.status_code == 200:
                    self._callback(True)
                else:
                    self._callback(False)
        except Exception as e:
            logger.exception(f"Error in upload: {e}")
            if self._callback:
                self._callback(False)  # delete data because it was corrupted

    @staticmethod
    def get_body(upload_string: str) -> str:
        return f"""
            {{
                "gateway": "python_webhook",
                "data": "{upload_string}",
                "published_at": "{datetime.datetime.now().isoformat()}"
            }}
            """

    @classmethod
    def upload(cls, data: dict[int, pd.DataFrame]) -> requests.Response:
        abs_time = int(datetime.datetime.now().timestamp())
        devices = []

        for sn, df in data.items():
            # make all inf values in df_p2_pro 0
            df = df.replace([float("inf"), -float("inf")], 0)

            # detect ms timestamp and convert to s
            if df.index[0] > 1e12:
                df.index = df.index / 1e3
                df.index = df.index.astype(int)

            devices.append(create_proto_device(sn, abs_time, df))

        combined_entry = create_combined_entry(devices=devices, abs_timestamp=abs_time)

        proto_str = combined_entry.SerializeToString()
        proto_str_base64 = base64.b64encode(proto_str).decode()

        body = cls.get_body(proto_str_base64)
        r = requests.post(cls.URL, headers=cls.HEADERS, data=body, timeout=10)
        return r


def read_pickle_file(file_path: str) -> dict[int, pd.DataFrame]:
    with open(file_path, "rb") as f:
        data = pickle.load(f)
    return data


if __name__ == "__main__":
    data = read_pickle_file("partector_data.pkl")

    uploader = NaneosUploadThread(
        data, callback=lambda success: print(f"Upload success: {success}")
    )
    uploader.start()
    uploader.join()



from datetime import datetime, timezone
from typing import Any, Optional

from naneos.logger.custom_logger import get_naneos_logger
from naneos.partector.blueprints._data_structure import (
    PARTECTOR2_PRO_CS_DATA_STRUCTURE_V315,
    NaneosDeviceDataPoint,
)
from naneos.partector.partector2_pro import Partector2Pro

logger = get_naneos_logger(__name__)


class Partector2ProCs(Partector2Pro):
    CS_OFF = 0
    CS_ON = 1
    CS_UNKNOWN = -1

    def __init__(
        self,
        serial_number: Optional[int] = None,
        port: Optional[str] = None,
        verb_freq: int = 1,
        **kwargs: Any,
    ) -> None:
        self._catalyst_state = self.CS_UNKNOWN
        self._auto_mode = True

        self._callback_catalyst = kwargs.get("callback_catalyst", None)
        if self._callback_catalyst is None:
            logger.error("No callback function for catalyst state given!")
            raise ValueError("No callback function for catalyst state given!")
        super().__init__(serial_number, port, verb_freq, "P2proCS")

    def _init_serial_data_structure(self) -> None:
        self.device_type = NaneosDeviceDataPoint.DEV_TYPE_P2PRO_CS
        self._data_structure = PARTECTOR2_PRO_CS_DATA_STRUCTURE_V315

    def _set_verbose_freq(self, freq: int) -> None:
        if freq == 0:
            self._write_line("X0000!")
        else:
            if self._fw >= 311:
                self._data_structure = PARTECTOR2_PRO_CS_DATA_STRUCTURE_V315

            self._write_line("h2001!")  # activates harmonics output
            self._write_line("M0004!")  # activates size dist mode
            self._write_line("X0006!")  # activates verbose mode

    def set_catalyst_state(self, state: str) -> None:
        """Sets the catalyst state to on, off or auto."""
        if not self._connected:
            return

        if state == "on":
            self._write_line("CSon!")
            self._cs_state = self.CS_ON
            self._auto_mode = False
        elif state == "off":
            self._write_line("CSoff!")
            self._cs_state = self.CS_OFF
            self._auto_mode = False
        elif state == "auto":
            self._write_line("CSauto!")
            self._auto_mode = True
        else:
            logger.warning(f"Unknown catalyst state: {state} -> nothing done.")
            return

        logger.info(f"Catalyst state set to {state}.")

    def _serial_reading_routine(self) -> None:
        line = self._read_line()

        if not line or line == "":
            return

        if "CS_on" in line:
            self._catalyst_state = self.CS_ON
            if self._callback_catalyst:
                self._callback_catalyst(True)
            self._mark_cs_change()
            return
        elif "CS_off" in line:
            self._catalyst_state = self.CS_OFF
            if self._callback_catalyst:
                self._callback_catalyst(False)
            self._mark_cs_change()
            return

        self._put_line_to_queue(line)

    def _mark_cs_change(self) -> None:
        try:
            last_line = self._queue.pop()
            last_line[-1] = f"1{last_line[-1]}"
            self._queue.append(last_line)
        except Exception as excep:
            logger.warning(f"Could not mark catalyst state change: {excep}")

    def _put_line_to_queue(self, line: str) -> None:
        unix_timestamp = int(datetime.now(tz=timezone.utc).timestamp())
        data = [unix_timestamp] + line.split("\t")

        self._notify_message_received()

        if len(data) != len(self._data_structure):
            self._queue_info.append(data)
            return

        state: Optional[int] = None
        try:
            state = int(data[-1])
        except Exception as excep:
            logger.warning(f"Could not parse catalyst state in backup function: {excep}")

        if state in [0, 1] and self._catalyst_state != state:
            if self._callback_catalyst:
                self._callback_catalyst(state == 1)
            self._catalyst_state = state
            logger.warning(f"Set catalyst state to {state} by backup function to.")

        self._queue.append(data)


if __name__ == "__main__":
    import time

    import pandas as pd

    def test_callback(state: bool) -> None:
        logger.info(f"Catalyst state changed to {state}.")

    logger.info("Starting...")

    p2 = Partector2ProCs(serial_number=8448, callback_catalyst=test_callback)

    start = time.time()
    df = pd.DataFrame()
    while time.time() - start < 30:
        data_points = p2.get_data()
        data: dict[int, pd.DataFrame] = {}
        for point in data_points:
            data = NaneosDeviceDataPoint.add_data_point_to_dict(data, point)
        df_tmp = next(iter(data.values()), pd.DataFrame())

        if len(df_tmp) > 0:
            df = pd.concat([df, df_tmp])
        time.sleep(0.01)

    print(df)
    # df.to_pickle("tests/df_garagae.pkl")

    print("Closing...")
    p2.close(blocking=True)
    print("Closed!")



import asyncio
import sys
import threading
import time

import pandas as pd
from bleak import BleakScanner
from bleak.backends.device import BLEDevice

from naneos.logger import LEVEL_WARNING, get_naneos_logger
from naneos.partector.blueprints._data_structure import (
    NaneosDeviceDataPoint,
)
from naneos.partector_ble.partector_ble_connection import PartectorBleConnection
from naneos.partector_ble.partector_ble_scanner import PartectorBleScanner

pd.set_option("future.no_silent_downcasting", True)

logger = get_naneos_logger(__name__, LEVEL_WARNING)


class PartectorBleManager(threading.Thread):
    def __init__(self) -> None:
        super().__init__(daemon=True)
        self._stop_event = threading.Event()
        self._task_stop_event = asyncio.Event()

        self._queue_scanner = PartectorBleScanner.create_scanner_queue()
        self._queue_connection = PartectorBleConnection.create_connection_queue()
        self._connections: dict[int, tuple[asyncio.Task, int]] = {}  # key: serial_number

        self._data: dict[int, pd.DataFrame] = {}

    def get_data(self) -> dict[int, pd.DataFrame]:
        """Returns the data dictionary and deletes it."""
        data = self._data
        self._data = {}
        return data

    def stop(self) -> None:
        self._task_stop_event.set()
        self._stop_event.set()

    def run(self) -> None:
        try:
            asyncio.run(self._async_run())
        except RuntimeError as e:
            logger.exception(f"BLEManager loop exited with: {e}")

    def get_connected_device_strings(self) -> list[str]:
        """Returns a list of connected device strings."""
        # first make a copy to avoid runtime dict change issues
        connections_copy = self._connections.copy()
        sns = connections_copy.keys()
        device_types = [connections_copy[s][1] for s in sns]

        sns_list = []
        for sn, dev_type in zip(sns, device_types):
            if dev_type == NaneosDeviceDataPoint.DEV_TYPE_P2PRO:
                sns_list.append(f"SN{sn} (P2 Pro)")
        for sn, dev_type in zip(sns, device_types):
            if dev_type == NaneosDeviceDataPoint.DEV_TYPE_P2:
                sns_list.append(f"SN{sn} (P2)")

        return sns_list

    def get_connected_serial_numbers(self) -> list[int | None]:
        """Returns a list of connected serial numbers."""
        return list(self._connections.keys())

    async def _bleak_is_bluetooth_adapter_available(self) -> bool:
        """Check if the Bluetooth adapter is available and powered on."""
        try:
            # Try to get adapter info - this will fail if adapter is not available
            scanner = BleakScanner()
            # Test if we can discover devices briefly
            await scanner.start()
            await scanner.stop()
            return True
        except Exception as e:
            logger.debug(f"Bluetooth adapter not available: {e}")
            return False

    async def _linux_is_bluetooth_adapter_available(self) -> bool:
        """
        Nutzt BlueZ (bluetoothctl show), um zu prüfen, ob
        - ein Bluetooth-Controller existiert und
        - er eingeschaltet ("Powered: yes") ist.
        """
        try:
            proc = await asyncio.create_subprocess_exec(
                "bluetoothctl",
                "show",
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            stdout, stderr = await proc.communicate()

            if proc.returncode != 0:
                logger.debug(
                    "bluetoothctl show failed with code %s: %s",
                    proc.returncode,
                    stderr.decode(errors="ignore").strip(),
                )
                return False

            output = stdout.decode(errors="ignore")

            if "No default controller available" in output:
                logger.debug("No default Bluetooth controller available (BlueZ).")
                return False

            powered = None
            for line in output.splitlines():
                line = line.strip()
                if line.lower().startswith("powered:"):
                    powered = "yes" in line.lower()
                    break

            if powered is not None:
                return powered

            logger.debug("Bluetooth controller found but no 'Powered' field in output.")
            return False

        except FileNotFoundError:
            logger.debug("bluetoothctl not found on system.")
            return False

        except Exception as e:
            logger.debug(f"Error while checking Bluetooth adapter via bluetoothctl: {e}")
            return False

    async def _is_bluetooth_adapter_available(self) -> bool:
        if sys.platform.startswith("linux"):
            return await self._linux_is_bluetooth_adapter_available()
        else:
            return await self._bleak_is_bluetooth_adapter_available()

    async def _wait_for_bluetooth_adapter(self) -> None:
        """Wait for the Bluetooth adapter to become available."""
        adapter_check_interval = 3.0  # seconds

        while not self._stop_event.is_set():
            if await self._is_bluetooth_adapter_available():
                logger.info("Bluetooth adapter is available and ready.")
                return

            logger.info(
                f"Bluetooth adapter not available. Retrying in {adapter_check_interval} seconds..."
            )
            await asyncio.sleep(adapter_check_interval)

    async def _async_run(self):
        self._loop = asyncio.get_event_loop()
        while not self._stop_event.is_set():
            try:
                # Wait for Bluetooth adapter to become available
                await self._wait_for_bluetooth_adapter()
                self._task_stop_event.clear()

                async with PartectorBleScanner(loop=self._loop, queue=self._queue_scanner):
                    logger.info("Scanner started.")
                    await self._manager_loop()
                await self._kill_all_connections()  # just to be safe
            except asyncio.CancelledError:
                logger.info("BLEManager cancelled.")
            finally:
                logger.info("BLEManager cleanup complete.")

    async def _manager_loop(self) -> None:
        while not self._stop_event.is_set():
            try:
                if not await self._is_bluetooth_adapter_available():
                    logger.warning("Bluetooth adapter lost. Stopping all connections...")
                    await self._kill_all_connections()
                    return

                await asyncio.sleep(0.3)

                await self._scanner_queue_routine()
                await self._connection_queue_routine()
                await self._check_device_types()
                await self._remove_done_tasks()

            except asyncio.TimeoutError:
                continue
            except Exception as e:
                logger.exception(f"Error in manager loop: {e}")

        await self._finish_all_connections()

    async def _kill_all_connections(self) -> None:
        self._task_stop_event.set()

        for serial in list(self._connections.keys()):
            if not self._connections[serial][0].done():
                logger.info(f"Cancelling connection task {serial}.")
                self._connections[serial][0].cancel()
            self._connections.pop(serial, None)
            logger.info(f"{serial}: Connection task cancelled and popped.")

    async def _finish_all_connections_blocking(self) -> None:
        while list(self._connections.keys()):
            serial = list(self._connections.keys())[0]

            if not self._connections[serial][0].done():
                await asyncio.sleep(1)
            else:
                self._connections.pop(serial, None)

    async def _finish_all_connections(self) -> None:
        self._task_stop_event.set()
        await asyncio.sleep(1)  # give tasks some time to finish gracefully

        # wait max 5s for _finish_all_connections_blocking to finish
        try:
            await asyncio.wait_for(self._finish_all_connections_blocking(), timeout=7)
        except asyncio.TimeoutError:
            logger.warning("Timeout waiting for connections to finish. Forcing cancellation.")

        for serial in list(self._connections.keys()):
            if not self._connections[serial][0].done():
                logger.warning(f"Forcing connection task {serial} to cancel.")
                self._connections[serial][0].cancel()
                await asyncio.sleep(0.1)  # small delay to allow cancellation to propagate
                # logger.info(f"Waiting for connection task {serial} to finish.")
                # await self._connections[serial]

            self._connections.pop(serial, None)
            logger.info(f"{serial}: Connection task finished and popped.")

    async def _task_connection(self, device: BLEDevice, serial: int) -> None:
        try:
            async with PartectorBleConnection(
                device=device, loop=self._loop, serial_number=serial, queue=self._queue_connection
            ):
                while not self._task_stop_event.is_set():
                    await asyncio.sleep(0.5)

        except asyncio.CancelledError:
            logger.info(f"{serial}: Connection task cancelled.")
        except Exception as e:
            logger.warning(f"{serial}: Connection task failed: {e}")
        finally:
            logger.info(f"{serial}: Connection task finished.")

    async def _scanner_queue_routine(self) -> None:
        """Process scanner queue with batch collection to reduce DataFrame operations.

        Instead of calling add_data_point_to_dict for each item (expensive),
        collect all items first, then add them in bulk.
        """
        to_check: dict[int, BLEDevice] = {}
        batch_data: list[NaneosDeviceDataPoint] = []

        # Collect all available items from queue (non-blocking batch)
        while not self._queue_scanner.empty():
            try:
                device, decoded = self._queue_scanner.get_nowait()
            except asyncio.QueueEmpty:
                break

            if not decoded.serial_number:
                continue

            batch_data.append(decoded)
            to_check[decoded.serial_number] = device

        # Add all data points at once (more efficient than individual additions)
        for decoded in batch_data:
            self._data = NaneosDeviceDataPoint.add_data_point_to_dict(self._data, decoded)

        # check for new devices
        for serial, device in to_check.items():
            if serial in self._connections:
                continue  # already connected

            logger.info(f"New device detected: serial={serial}, address={device.address}")
            task = self._loop.create_task(self._task_connection(device, serial))
            self._connections[serial] = (task, NaneosDeviceDataPoint.DEV_TYPE_P2)

    async def _connection_queue_routine(self) -> None:
        """Process connection queue with batch collection to reduce DataFrame operations.

        Instead of calling add_data_point_to_dict for each item (expensive),
        collect all items first, then add them in bulk.
        """
        batch_data: list[NaneosDeviceDataPoint] = []

        # Collect all available items from queue (non-blocking batch)
        while not self._queue_connection.empty():
            try:
                data = self._queue_connection.get_nowait()
            except asyncio.QueueEmpty:
                break

            batch_data.append(data)

        # Add all data points at once (more efficient than individual additions)
        for data in batch_data:
            self._data = NaneosDeviceDataPoint.add_data_point_to_dict(self._data, data)

    async def _check_device_types(self) -> None:
        for serial in self._data.keys():
            if serial not in self._connections:
                continue

            current_type = self._connections[serial][1]
            data_points = self._data[serial]

            # get last value of device_type column
            if data_points.empty:
                continue
            last_device_type = data_points["device_type"].iloc[-1]
            if last_device_type != current_type:
                self._connections[serial] = (
                    self._connections[serial][0],
                    last_device_type,
                )

    async def _remove_done_tasks(self) -> None:
        """Remove completed tasks from the connections dictionary."""
        for serial in list(self._connections.keys()):
            if self._connections[serial][0].done():
                self._connections.pop(serial, None)
                logger.info(f"{serial}: Connection task finished and popped.")


if __name__ == "__main__":
    manager = PartectorBleManager()
    manager.start()

    for _ in range(2):
        time.sleep(10)  # Allow some time for the scanner to start
        data = manager.get_data()

        print(f"Connected serial numbers: {manager.get_connected_serial_numbers()}")
        print("Collected data:")
        print()

        for sn, df in data.items():
            print(f"SN: {sn}")
            print(df)
            print("-" * 40)
            print()

    manager.stop()
    manager.join()



from typing import Optional

from naneos.partector.blueprints._data_structure import NaneosDeviceDataPoint
from naneos.partector_ble.decoder.partector_ble_decoder_blueprint import (
    PartectorBleDecoderBlueprint,
)


class PartectorBleDecoderAuxError(PartectorBleDecoderBlueprint):
    """
    Decode the std advertisement data from the Partector device.
    """

    OFFSET_ERROR = slice(2, 6)
    OFFSET_IDIFF_DELAY_ON = slice(6, 7)
    OFFSET_IDIFF_DELAY_OFF = slice(7, 8)
    OFFSET_IDIFF_AVERAGE = slice(8, 10)
    OFFSET_IDIFF_STDV = slice(10, 12)
    OFFSET_IDIFF_MAX = slice(12, 14)
    OFFSET_UCOR_ONSET = slice(14, 16)

    FACTOR_DIFFUSION_CURRENT = 0.01

    # == External used methods =====================================================================
    @classmethod
    def decode(
        cls, data: bytes, data_structure: Optional[NaneosDeviceDataPoint] = None
    ) -> NaneosDeviceDataPoint:
        """
        Decode the auxiliary characteristic data from the Partector device.
        """
        decoded_data = NaneosDeviceDataPoint(
            device_status=cls._get_error(data),
            diffusion_current_delay_on=cls._get_idiff_delay_on(data),
            diffusion_current_delay_off=cls._get_idiff_delay_off(data),
            diffusion_current_average=cls._get_idiff_average(data),
            diffusion_current_stddev=cls._get_idiff_stddev(data),
            diffusion_current_max=cls._get_idiff_max(data),
            corona_voltage_onset=cls._get_ucor_onset(data),
        )

        if not data_structure:
            return decoded_data

        # Fill the given data_structure with the decoded data
        for field in NaneosDeviceDataPoint.BLE_AUX_ERROR_FIELD_NAMES:
            setattr(data_structure, field, getattr(decoded_data, field))
        return data_structure

    # == Helpers ===================================================================================
    @classmethod
    def _get_error(cls, data: bytes) -> int:
        """
        Get the error bits from the advertisement data.
        """
        val = int.from_bytes(data[cls.OFFSET_ERROR], byteorder="little")
        return val

    @classmethod
    def _get_idiff_delay_on(cls, data: bytes) -> int:
        """
        Get the diffusion current delay on from the advertisement data.
        """
        val = int.from_bytes(data[cls.OFFSET_IDIFF_DELAY_ON], byteorder="little")
        return val

    @classmethod
    def _get_idiff_delay_off(cls, data: bytes) -> int:
        """
        Get the diffusion current delay off from the advertisement data.
        """
        val = int.from_bytes(data[cls.OFFSET_IDIFF_DELAY_OFF], byteorder="little")
        return val

    @classmethod
    def _get_idiff_average(cls, data: bytes) -> float:
        """
        Get the diffusion current average from the advertisement data.
        """
        val = int.from_bytes(data[cls.OFFSET_IDIFF_AVERAGE], byteorder="little")
        return val * cls.FACTOR_DIFFUSION_CURRENT

    @classmethod
    def _get_idiff_stddev(cls, data: bytes) -> float:
        """
        Get the diffusion current standard deviation from the advertisement data.
        """
        val = int.from_bytes(data[cls.OFFSET_IDIFF_STDV], byteorder="little")
        return val * cls.FACTOR_DIFFUSION_CURRENT

    @classmethod
    def _get_idiff_max(cls, data: bytes) -> float:
        """
        Get the diffusion current maximum from the advertisement data.
        """
        val = int.from_bytes(data[cls.OFFSET_IDIFF_MAX], byteorder="little")
        return val * cls.FACTOR_DIFFUSION_CURRENT

    @classmethod
    def _get_ucor_onset(cls, data: bytes) -> int:
        """
        Get the corona voltage onset from the advertisement data.
        """
        val = int.from_bytes(data[cls.OFFSET_UCOR_ONSET], byteorder="little")
        return val



#!/usr/bin/env python3
"""
Demo script to demonstrate Bluetooth adapter checking functionality.
This script shows how the PartectorBleManager handles Bluetooth adapter availability.
"""

import asyncio
import time

from naneos.partector_ble.partector_ble_manager import PartectorBleManager


async def demo_adapter_checking():
    """Demonstrate the Bluetooth adapter checking functionality."""
    print("=== Bluetooth Adapter Checking Demo ===")
    print()

    manager = PartectorBleManager()

    # Test the adapter checking method directly
    print("Testing Bluetooth adapter availability...")
    is_available = await manager._is_bluetooth_adapter_available()
    print(f"Bluetooth adapter available: {is_available}")
    print()

    if is_available:
        print("Starting BLE Manager (it will automatically check adapter status)...")
        manager.start()

        # Let it run for a short time
        time.sleep(5)

        print(f"Connected devices: {manager.get_connected_device_strings()}")
        print("Stopping manager...")
        manager.stop()
        manager.join()
        print("Manager stopped.")
    else:
        print("Bluetooth adapter not available. Manager would wait until it becomes available.")


if __name__ == "__main__":
    asyncio.run(demo_adapter_checking())



import sys
from pathlib import Path

import serial
import serial.tools.list_ports as ls


def list_serial_ports(ports_exclude: list = []) -> list[str]:
    """Returns a list of serial ports available on the system.

    Raises:
        OSError: If the platform is not supported.

    Returns:
        list[str]: A list of serial ports available on the system.
    """
    # ports: list[str] = _get_all_open_ports()
    ports: list[str] = _get_all_dosemet_ports(ports_exclude)
    ports = _check_port_function(ports)

    return ports


def _get_all_dosemet_ports(ports_exclude: list) -> list[str]:
    ports: list[str] = []

    all_ports = ls.comports()
    all_ports = [port for port in all_ports if port.device not in ports_exclude]
    for port in all_ports:
        if (port.pid == 5 and port.vid == 65535) or (
            port.serial_number and "dosemet" in port.serial_number.lower()
        ):
            ports.append(port.device)

    return ports


def _get_all_open_ports() -> list[str]:
    ports: list[str] = []

    if sys.platform.startswith("win"):
        ports = ["COM%s" % (i + 1) for i in range(256)]
    elif sys.platform.startswith("linux") or sys.platform.startswith("cygwin"):
        ports = [str(port) for port in Path("/dev").glob("tty[A-Za-z]*")]
    elif sys.platform.startswith("darwin"):  # mac
        ports = [str(port) for port in Path("/dev").glob("tty.*") if "Bluetooth" not in str(port)]
    else:
        raise OSError(f"Unsupported platform: {sys.platform}")

    return ports


def _check_port_function(ports: list[str]) -> list[str]:
    working_ports: list[str] = []

    for port in ports:
        for _ in range(100):  # I have to do this because of windows and P2's 100Hz output
            try:
                s = serial.Serial(port)
                s.write(b"X0000!")
                s.close()
                working_ports.append(port)
                break
            except (OSError, serial.SerialException):
                pass

    return working_ports



import asyncio
from typing import Optional

from bleak.backends.device import BLEDevice

from naneos.partector.blueprints._data_structure import NaneosDeviceDataPoint
from naneos.partector_ble.partector_ble_connection import PartectorBleConnection
from naneos.partector_ble.partector_ble_scanner import PartectorBleScanner

SNS = {8617}  # serial numbers to connect to for testing


async def _map_sn_to_device(
    queue: asyncio.Queue[tuple[BLEDevice, NaneosDeviceDataPoint]],
) -> Optional[dict[int, BLEDevice]]:
    device_dict = {}
    while not queue.empty():
        device, data = await queue.get()
        if data.serial_number:
            device_dict[data.serial_number] = device

    if not device_dict:
        return None

    return device_dict


async def async_test_connection(with_context_manager: bool) -> None:
    """Helper function to test the scanner."""
    loop = asyncio.get_event_loop()
    queue_scanner = PartectorBleScanner.create_scanner_queue()
    queue_connection = PartectorBleConnection.create_connection_queue()
    conn_list = []  # serial number to connection mapping

    if with_context_manager:
        async with PartectorBleScanner(loop=loop, queue=queue_scanner) as scanner:
            await asyncio.sleep(5)
    else:
        scanner = PartectorBleScanner(loop=loop, queue=queue_scanner)
        scanner.start()
        try:
            await asyncio.sleep(5)  # Asynchronous sleep to allow the loop to run
        finally:
            await scanner.stop()

    device_dict = await _map_sn_to_device(queue_scanner)
    assert isinstance(device_dict, dict)

    device_dict = {k: v for k, v in device_dict.items() if k in SNS}
    assert len(device_dict) == len(SNS)

    # start connections for all devices
    for serial_number, device in device_dict.items():
        conn_list.append(
            PartectorBleConnection(
                device=device, loop=loop, serial_number=serial_number, queue=queue_connection
            )
        )
        conn_list[-1].start()

    await asyncio.sleep(5)

    # stop connections for all devices
    for conn in conn_list:
        await conn.stop()


def test_connection() -> None:
    """Test the scanner functionality."""
    asyncio.run(async_test_connection(with_context_manager=False))


def test_connection_with_context_manager() -> None:
    """Test the scanner functionality with context manager."""
    asyncio.run(async_test_connection(with_context_manager=True))



import threading
import time

import pandas as pd

from naneos.logger import LEVEL_WARNING, get_naneos_logger
from naneos.partector.blueprints._data_structure import NaneosDeviceDataPoint
from naneos.partector.partector1 import Partector1
from naneos.partector.partector2 import Partector2
from naneos.partector.partector2_pro import Partector2Pro
from naneos.partector.scanPartector import scan_for_serial_partectors

logger = get_naneos_logger(__name__, LEVEL_WARNING)


class PartectorSerialManager(threading.Thread):
    def __init__(self) -> None:
        super().__init__(daemon=True)
        self._stop_event = threading.Event()

        self._data: dict[int, pd.DataFrame] = {}

        self._connected_p1: dict[str, Partector1] = {}
        self._connected_p2: dict[str, Partector2] = {}
        self._connected_p2_pro: dict[str, Partector2Pro] = {}

    def get_data(self) -> dict[int, pd.DataFrame]:
        """Fetches the data from all connected devices and returns it."""
        self._fetch_data()
        data = self._data
        self._data = {}
        return data

    def _fetch_data(self):
        """Returns the data dictionary and deletes it."""
        for port in list(self._connected_p1.keys()):
            points = self._connected_p1[port].get_data()
            for point in points:
                self._data = NaneosDeviceDataPoint.add_data_point_to_dict(self._data, point)

        for port in list(self._connected_p2.keys()):
            points = self._connected_p2[port].get_data()
            for point in points:
                self._data = NaneosDeviceDataPoint.add_data_point_to_dict(self._data, point)

        for port in list(self._connected_p2_pro.keys()):
            points = self._connected_p2_pro[port].get_data()
            for point in points:
                self._data = NaneosDeviceDataPoint.add_data_point_to_dict(self._data, point)

    def stop(self) -> None:
        self._stop_event.set()

    def run(self) -> None:
        try:
            self._manager_loop()
        except RuntimeError as e:
            logger.exception(f"SerialManager loop exited with: {e}")

    def get_connected_device_strings(self) -> list[str]:
        """Returns a list of connected device strings."""
        p1_strings = [f"SN{p._sn} (P1)" for port, p in self._connected_p1.items()]
        p2_strings = [f"SN{p._sn} (P2)" for port, p in self._connected_p2.items()]
        p2_pro_strings = [f"SN{p._sn} (P2 Pro)" for port, p in self._connected_p2_pro.items()]

        return p1_strings + p2_strings + p2_pro_strings

    def get_gain_test_activating_devices(self) -> list[int | None]:
        """Returns a list of serial numbers of devices with gain test active."""
        # _wait_with_data_output_until is larger than time.time()
        p2_gain_test = [
            p._sn
            for p in self._connected_p2.values()
            if p._wait_with_data_output_until > time.time()
        ]
        p2_pro_gain_test = [
            p._sn
            for p in self._connected_p2_pro.values()
            if p._wait_with_data_output_until > time.time()
        ]

        return p2_gain_test + p2_pro_gain_test

    def get_connected_addresses(self) -> list[str]:
        p1_ports = list(self._connected_p1.keys())
        p2_ports = list(self._connected_p2.keys())
        p2_pro_ports = list(self._connected_p2_pro.keys())

        return p1_ports + p2_ports + p2_pro_ports

    def get_connected_serial_numbers(self) -> list[int | None]:
        p1_serials = [p._sn for p in self._connected_p1.values()]
        p2_serials = [p._sn for p in self._connected_p2.values()]
        p2_pro_serials = [p._sn for p in self._connected_p2_pro.values()]

        return p1_serials + p2_serials + p2_pro_serials

    def _manager_loop(self) -> None:
        while not self._stop_event.is_set():
            try:
                possible_ports = scan_for_serial_partectors(
                    ports_exclude=self.get_connected_addresses()
                )

                self._disconnect_unplugged_ports()
                self._connect_to_new_ports(possible_ports)

                self._fetch_data()  # Fetch data from all connected devices

                time.sleep(1.0)  # Sleep to avoid busy waiting

            except Exception as e:
                logger.exception(f"Error in serial manager loop: {e}")

        self._close_all_ports()

    def _disconnect_unplugged_ports(self) -> None:
        """Disconnects all ports that are not in the possible_ports dictionary."""
        # Disconnect P1 ports
        for port in list(self._connected_p1.keys()):
            if not self._connected_p1[port]._connected:
                self._connected_p1[port].close()
                self._connected_p1.pop(port, None)

        # Disconnect P2 ports
        for port in list(self._connected_p2.keys()):
            if not self._connected_p2[port]._connected:
                self._connected_p2[port].close()
                self._connected_p2.pop(port, None)

        # Disconnect P2 Pro ports
        for port in list(self._connected_p2_pro.keys()):
            if not self._connected_p2_pro[port]._connected:
                print(f"Disconnecting P2 Pro port: {port}")
                self._connected_p2_pro[port].close()
                self._connected_p2_pro.pop(port, None)

    def _connect_to_new_ports(self, possible_ports: dict[str, dict[int, str]]) -> None:
        p1_ports = possible_ports["P1"].values()
        p2_ports = possible_ports["P2"].values()
        p2pro_ports = possible_ports["P2pro"].values()

        for port in p1_ports:
            self._connected_p1[port] = Partector1(port=port)
        for port in p2_ports:
            self._connected_p2[port] = Partector2(port=port)
        for port in p2pro_ports:
            self._connected_p2_pro[port] = Partector2Pro(port=port)

    def _close_all_ports(self) -> None:
        for port in list(self._connected_p1.keys()):
            self._connected_p1[port].close()
            self._connected_p1.pop(port, None)

        for port in list(self._connected_p2.keys()):
            self._connected_p2[port].close()
            self._connected_p2.pop(port, None)

        for port in list(self._connected_p2_pro.keys()):
            self._connected_p2_pro[port].close()
            self._connected_p2_pro.pop(port, None)


if __name__ == "__main__":
    manager = PartectorSerialManager()
    manager.start()

    for _ in range(20):
        time.sleep(10)  # Let the manager run for a while
        data = manager.get_data()

        print(f"Connected ports: {manager.get_connected_addresses()}")
        print(f"Connected serial numbers: {manager.get_connected_serial_numbers()}")
        print("Collected data:")
        print()

        for sn, df in data.items():
            print(f"SN: {sn}")
            print(df)
            print("-" * 40)
            print()

    manager.stop()
    manager.join()



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class CombinedData(_message.Message):
    __slots__ = ("abs_timestamp", "devices", "gateway_points_legacy", "position_points", "wind_points", "uplink")
    ABS_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    DEVICES_FIELD_NUMBER: _ClassVar[int]
    GATEWAY_POINTS_LEGACY_FIELD_NUMBER: _ClassVar[int]
    POSITION_POINTS_FIELD_NUMBER: _ClassVar[int]
    WIND_POINTS_FIELD_NUMBER: _ClassVar[int]
    UPLINK_FIELD_NUMBER: _ClassVar[int]
    abs_timestamp: int
    devices: _containers.RepeatedCompositeFieldContainer[Device]
    gateway_points_legacy: _containers.RepeatedCompositeFieldContainer[GatewayPointLegacy]
    position_points: _containers.RepeatedCompositeFieldContainer[PositionPoint]
    wind_points: _containers.RepeatedCompositeFieldContainer[WindPoint]
    uplink: Uplink
    def __init__(self, abs_timestamp: _Optional[int] = ..., devices: _Optional[_Iterable[_Union[Device, _Mapping]]] = ..., gateway_points_legacy: _Optional[_Iterable[_Union[GatewayPointLegacy, _Mapping]]] = ..., position_points: _Optional[_Iterable[_Union[PositionPoint, _Mapping]]] = ..., wind_points: _Optional[_Iterable[_Union[WindPoint, _Mapping]]] = ..., uplink: _Optional[_Union[Uplink, _Mapping]] = ...) -> None: ...

class Device(_message.Message):
    __slots__ = ("type", "serial_number", "device_points", "ui_curve")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    SERIAL_NUMBER_FIELD_NUMBER: _ClassVar[int]
    DEVICE_POINTS_FIELD_NUMBER: _ClassVar[int]
    UI_CURVE_FIELD_NUMBER: _ClassVar[int]
    type: int
    serial_number: int
    device_points: _containers.RepeatedCompositeFieldContainer[DevicePoint]
    ui_curve: _containers.RepeatedCompositeFieldContainer[UiCurve]
    def __init__(self, type: _Optional[int] = ..., serial_number: _Optional[int] = ..., device_points: _Optional[_Iterable[_Union[DevicePoint, _Mapping]]] = ..., ui_curve: _Optional[_Iterable[_Union[UiCurve, _Mapping]]] = ...) -> None: ...

class DevicePoint(_message.Message):
    __slots__ = ("timestamp", "device_status", "ldsa", "average_particle_diameter", "particle_number_concentration", "temperature", "relative_humidity", "battery_voltage", "particle_mass", "corona_voltage", "diffusion_current", "deposition_voltage", "flow", "ambient_pressure", "electrometer_1_offset", "electrometer_2_offset", "electrometer_1_gain", "electrometer_2_gain", "diffusion_current_offset", "particle_number_10nm", "particle_number_16nm", "particle_number_26nm", "particle_number_43nm", "particle_number_70nm", "particle_number_114nm", "particle_number_185nm", "particle_number_300nm", "particle_surface", "sigma_size_dist", "steps_inversion", "current_dist_0", "current_dist_1", "current_dist_2", "current_dist_3", "current_dist_4", "pump_current", "pump_pwm", "cs_status", "diffusion_current_stddev", "diffusion_current_delay_on", "diffusion_current_delay_off", "pump_voltage", "differential_pressure", "channel_pressure", "supply_voltage_5V", "positive_voltage_3V3", "negative_voltage_3V3", "usb_cc_voltage", "firmware_version")
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    DEVICE_STATUS_FIELD_NUMBER: _ClassVar[int]
    LDSA_FIELD_NUMBER: _ClassVar[int]
    AVERAGE_PARTICLE_DIAMETER_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_NUMBER_CONCENTRATION_FIELD_NUMBER: _ClassVar[int]
    TEMPERATURE_FIELD_NUMBER: _ClassVar[int]
    RELATIVE_HUMIDITY_FIELD_NUMBER: _ClassVar[int]
    BATTERY_VOLTAGE_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_MASS_FIELD_NUMBER: _ClassVar[int]
    CORONA_VOLTAGE_FIELD_NUMBER: _ClassVar[int]
    DIFFUSION_CURRENT_FIELD_NUMBER: _ClassVar[int]
    DEPOSITION_VOLTAGE_FIELD_NUMBER: _ClassVar[int]
    FLOW_FIELD_NUMBER: _ClassVar[int]
    AMBIENT_PRESSURE_FIELD_NUMBER: _ClassVar[int]
    ELECTROMETER_1_OFFSET_FIELD_NUMBER: _ClassVar[int]
    ELECTROMETER_2_OFFSET_FIELD_NUMBER: _ClassVar[int]
    ELECTROMETER_1_GAIN_FIELD_NUMBER: _ClassVar[int]
    ELECTROMETER_2_GAIN_FIELD_NUMBER: _ClassVar[int]
    DIFFUSION_CURRENT_OFFSET_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_NUMBER_10NM_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_NUMBER_16NM_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_NUMBER_26NM_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_NUMBER_43NM_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_NUMBER_70NM_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_NUMBER_114NM_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_NUMBER_185NM_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_NUMBER_300NM_FIELD_NUMBER: _ClassVar[int]
    PARTICLE_SURFACE_FIELD_NUMBER: _ClassVar[int]
    SIGMA_SIZE_DIST_FIELD_NUMBER: _ClassVar[int]
    STEPS_INVERSION_FIELD_NUMBER: _ClassVar[int]
    CURRENT_DIST_0_FIELD_NUMBER: _ClassVar[int]
    CURRENT_DIST_1_FIELD_NUMBER: _ClassVar[int]
    CURRENT_DIST_2_FIELD_NUMBER: _ClassVar[int]
    CURRENT_DIST_3_FIELD_NUMBER: _ClassVar[int]
    CURRENT_DIST_4_FIELD_NUMBER: _ClassVar[int]
    PUMP_CURRENT_FIELD_NUMBER: _ClassVar[int]
    PUMP_PWM_FIELD_NUMBER: _ClassVar[int]
    CS_STATUS_FIELD_NUMBER: _ClassVar[int]
    DIFFUSION_CURRENT_STDDEV_FIELD_NUMBER: _ClassVar[int]
    DIFFUSION_CURRENT_DELAY_ON_FIELD_NUMBER: _ClassVar[int]
    DIFFUSION_CURRENT_DELAY_OFF_FIELD_NUMBER: _ClassVar[int]
    PUMP_VOLTAGE_FIELD_NUMBER: _ClassVar[int]
    DIFFERENTIAL_PRESSURE_FIELD_NUMBER: _ClassVar[int]
    CHANNEL_PRESSURE_FIELD_NUMBER: _ClassVar[int]
    SUPPLY_VOLTAGE_5V_FIELD_NUMBER: _ClassVar[int]
    POSITIVE_VOLTAGE_3V3_FIELD_NUMBER: _ClassVar[int]
    NEGATIVE_VOLTAGE_3V3_FIELD_NUMBER: _ClassVar[int]
    USB_CC_VOLTAGE_FIELD_NUMBER: _ClassVar[int]
    FIRMWARE_VERSION_FIELD_NUMBER: _ClassVar[int]
    timestamp: int
    device_status: int
    ldsa: int
    average_particle_diameter: int
    particle_number_concentration: int
    temperature: int
    relative_humidity: int
    battery_voltage: int
    particle_mass: int
    corona_voltage: int
    diffusion_current: int
    deposition_voltage: int
    flow: int
    ambient_pressure: int
    electrometer_1_offset: int
    electrometer_2_offset: int
    electrometer_1_gain: int
    electrometer_2_gain: int
    diffusion_current_offset: int
    particle_number_10nm: int
    particle_number_16nm: int
    particle_number_26nm: int
    particle_number_43nm: int
    particle_number_70nm: int
    particle_number_114nm: int
    particle_number_185nm: int
    particle_number_300nm: int
    particle_surface: int
    sigma_size_dist: int
    steps_inversion: int
    current_dist_0: int
    current_dist_1: int
    current_dist_2: int
    current_dist_3: int
    current_dist_4: int
    pump_current: int
    pump_pwm: int
    cs_status: int
    diffusion_current_stddev: int
    diffusion_current_delay_on: int
    diffusion_current_delay_off: int
    pump_voltage: int
    differential_pressure: int
    channel_pressure: int
    supply_voltage_5V: int
    positive_voltage_3V3: int
    negative_voltage_3V3: int
    usb_cc_voltage: int
    firmware_version: int
    def __init__(self, timestamp: _Optional[int] = ..., device_status: _Optional[int] = ..., ldsa: _Optional[int] = ..., average_particle_diameter: _Optional[int] = ..., particle_number_concentration: _Optional[int] = ..., temperature: _Optional[int] = ..., relative_humidity: _Optional[int] = ..., battery_voltage: _Optional[int] = ..., particle_mass: _Optional[int] = ..., corona_voltage: _Optional[int] = ..., diffusion_current: _Optional[int] = ..., deposition_voltage: _Optional[int] = ..., flow: _Optional[int] = ..., ambient_pressure: _Optional[int] = ..., electrometer_1_offset: _Optional[int] = ..., electrometer_2_offset: _Optional[int] = ..., electrometer_1_gain: _Optional[int] = ..., electrometer_2_gain: _Optional[int] = ..., diffusion_current_offset: _Optional[int] = ..., particle_number_10nm: _Optional[int] = ..., particle_number_16nm: _Optional[int] = ..., particle_number_26nm: _Optional[int] = ..., particle_number_43nm: _Optional[int] = ..., particle_number_70nm: _Optional[int] = ..., particle_number_114nm: _Optional[int] = ..., particle_number_185nm: _Optional[int] = ..., particle_number_300nm: _Optional[int] = ..., particle_surface: _Optional[int] = ..., sigma_size_dist: _Optional[int] = ..., steps_inversion: _Optional[int] = ..., current_dist_0: _Optional[int] = ..., current_dist_1: _Optional[int] = ..., current_dist_2: _Optional[int] = ..., current_dist_3: _Optional[int] = ..., current_dist_4: _Optional[int] = ..., pump_current: _Optional[int] = ..., pump_pwm: _Optional[int] = ..., cs_status: _Optional[int] = ..., diffusion_current_stddev: _Optional[int] = ..., diffusion_current_delay_on: _Optional[int] = ..., diffusion_current_delay_off: _Optional[int] = ..., pump_voltage: _Optional[int] = ..., differential_pressure: _Optional[int] = ..., channel_pressure: _Optional[int] = ..., supply_voltage_5V: _Optional[int] = ..., positive_voltage_3V3: _Optional[int] = ..., negative_voltage_3V3: _Optional[int] = ..., usb_cc_voltage: _Optional[int] = ..., firmware_version: _Optional[int] = ...) -> None: ...

class UiCurve(_message.Message):
    __slots__ = ("timestamp", "U_values", "I_values")
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    U_VALUES_FIELD_NUMBER: _ClassVar[int]
    I_VALUES_FIELD_NUMBER: _ClassVar[int]
    timestamp: int
    U_values: _containers.RepeatedScalarFieldContainer[int]
    I_values: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, timestamp: _Optional[int] = ..., U_values: _Optional[_Iterable[int]] = ..., I_values: _Optional[_Iterable[int]] = ...) -> None: ...

class Uplink(_message.Message):
    __slots__ = ("type", "serial_number", "uplink_points")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    SERIAL_NUMBER_FIELD_NUMBER: _ClassVar[int]
    UPLINK_POINTS_FIELD_NUMBER: _ClassVar[int]
    type: int
    serial_number: int
    uplink_points: _containers.RepeatedCompositeFieldContainer[UplinkPoint]
    def __init__(self, type: _Optional[int] = ..., serial_number: _Optional[int] = ..., uplink_points: _Optional[_Iterable[_Union[UplinkPoint, _Mapping]]] = ...) -> None: ...

class UplinkPoint(_message.Message):
    __slots__ = ("timestamp", "firmware_version", "device_status", "cellular_signal", "battery_int_voltage", "voltage_in", "usb_voltage_out", "usb_cc_voltage", "temperature_int", "relative_humidity_int", "temperature_ext_1", "relative_humidity_ext_1", "temperature_ext_2", "relative_humidity_ext_2", "sraw_voc", "voc_index", "ambient_pressure", "co2_ppm_1", "co2_ppm_2", "pm1_0_mass", "pm2_5_mass", "pm4_0_mass", "pm10_0_mass", "water_level_adc_percent")
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    FIRMWARE_VERSION_FIELD_NUMBER: _ClassVar[int]
    DEVICE_STATUS_FIELD_NUMBER: _ClassVar[int]
    CELLULAR_SIGNAL_FIELD_NUMBER: _ClassVar[int]
    BATTERY_INT_VOLTAGE_FIELD_NUMBER: _ClassVar[int]
    VOLTAGE_IN_FIELD_NUMBER: _ClassVar[int]
    USB_VOLTAGE_OUT_FIELD_NUMBER: _ClassVar[int]
    USB_CC_VOLTAGE_FIELD_NUMBER: _ClassVar[int]
    TEMPERATURE_INT_FIELD_NUMBER: _ClassVar[int]
    RELATIVE_HUMIDITY_INT_FIELD_NUMBER: _ClassVar[int]
    TEMPERATURE_EXT_1_FIELD_NUMBER: _ClassVar[int]
    RELATIVE_HUMIDITY_EXT_1_FIELD_NUMBER: _ClassVar[int]
    TEMPERATURE_EXT_2_FIELD_NUMBER: _ClassVar[int]
    RELATIVE_HUMIDITY_EXT_2_FIELD_NUMBER: _ClassVar[int]
    SRAW_VOC_FIELD_NUMBER: _ClassVar[int]
    VOC_INDEX_FIELD_NUMBER: _ClassVar[int]
    AMBIENT_PRESSURE_FIELD_NUMBER: _ClassVar[int]
    CO2_PPM_1_FIELD_NUMBER: _ClassVar[int]
    CO2_PPM_2_FIELD_NUMBER: _ClassVar[int]
    PM1_0_MASS_FIELD_NUMBER: _ClassVar[int]
    PM2_5_MASS_FIELD_NUMBER: _ClassVar[int]
    PM4_0_MASS_FIELD_NUMBER: _ClassVar[int]
    PM10_0_MASS_FIELD_NUMBER: _ClassVar[int]
    WATER_LEVEL_ADC_PERCENT_FIELD_NUMBER: _ClassVar[int]
    timestamp: int
    firmware_version: int
    device_status: int
    cellular_signal: int
    battery_int_voltage: int
    voltage_in: int
    usb_voltage_out: int
    usb_cc_voltage: int
    temperature_int: int
    relative_humidity_int: int
    temperature_ext_1: int
    relative_humidity_ext_1: int
    temperature_ext_2: int
    relative_humidity_ext_2: int
    sraw_voc: int
    voc_index: int
    ambient_pressure: int
    co2_ppm_1: int
    co2_ppm_2: int
    pm1_0_mass: int
    pm2_5_mass: int
    pm4_0_mass: int
    pm10_0_mass: int
    water_level_adc_percent: int
    def __init__(self, timestamp: _Optional[int] = ..., firmware_version: _Optional[int] = ..., device_status: _Optional[int] = ..., cellular_signal: _Optional[int] = ..., battery_int_voltage: _Optional[int] = ..., voltage_in: _Optional[int] = ..., usb_voltage_out: _Optional[int] = ..., usb_cc_voltage: _Optional[int] = ..., temperature_int: _Optional[int] = ..., relative_humidity_int: _Optional[int] = ..., temperature_ext_1: _Optional[int] = ..., relative_humidity_ext_1: _Optional[int] = ..., temperature_ext_2: _Optional[int] = ..., relative_humidity_ext_2: _Optional[int] = ..., sraw_voc: _Optional[int] = ..., voc_index: _Optional[int] = ..., ambient_pressure: _Optional[int] = ..., co2_ppm_1: _Optional[int] = ..., co2_ppm_2: _Optional[int] = ..., pm1_0_mass: _Optional[int] = ..., pm2_5_mass: _Optional[int] = ..., pm4_0_mass: _Optional[int] = ..., pm10_0_mass: _Optional[int] = ..., water_level_adc_percent: _Optional[int] = ...) -> None: ...

class GatewayPointLegacy(_message.Message):
    __slots__ = ("timestamp", "serial_number", "firmware_version", "free_memory", "free_heap", "largest_free_block_heap", "cellular_signal", "battery_int_soc", "battery_ext_soc", "battery_ext_voltage", "charging_ext_voltage")
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    SERIAL_NUMBER_FIELD_NUMBER: _ClassVar[int]
    FIRMWARE_VERSION_FIELD_NUMBER: _ClassVar[int]
    FREE_MEMORY_FIELD_NUMBER: _ClassVar[int]
    FREE_HEAP_FIELD_NUMBER: _ClassVar[int]
    LARGEST_FREE_BLOCK_HEAP_FIELD_NUMBER: _ClassVar[int]
    CELLULAR_SIGNAL_FIELD_NUMBER: _ClassVar[int]
    BATTERY_INT_SOC_FIELD_NUMBER: _ClassVar[int]
    BATTERY_EXT_SOC_FIELD_NUMBER: _ClassVar[int]
    BATTERY_EXT_VOLTAGE_FIELD_NUMBER: _ClassVar[int]
    CHARGING_EXT_VOLTAGE_FIELD_NUMBER: _ClassVar[int]
    timestamp: int
    serial_number: int
    firmware_version: int
    free_memory: int
    free_heap: int
    largest_free_block_heap: int
    cellular_signal: int
    battery_int_soc: int
    battery_ext_soc: int
    battery_ext_voltage: int
    charging_ext_voltage: int
    def __init__(self, timestamp: _Optional[int] = ..., serial_number: _Optional[int] = ..., firmware_version: _Optional[int] = ..., free_memory: _Optional[int] = ..., free_heap: _Optional[int] = ..., largest_free_block_heap: _Optional[int] = ..., cellular_signal: _Optional[int] = ..., battery_int_soc: _Optional[int] = ..., battery_ext_soc: _Optional[int] = ..., battery_ext_voltage: _Optional[int] = ..., charging_ext_voltage: _Optional[int] = ...) -> None: ...

class PositionPoint(_message.Message):
    __slots__ = ("timestamp", "latitude", "longitude")
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    LATITUDE_FIELD_NUMBER: _ClassVar[int]
    LONGITUDE_FIELD_NUMBER: _ClassVar[int]
    timestamp: int
    latitude: float
    longitude: float
    def __init__(self, timestamp: _Optional[int] = ..., latitude: _Optional[float] = ..., longitude: _Optional[float] = ...) -> None: ...

class WindPoint(_message.Message):
    __slots__ = ("timestamp", "wind_speed", "wind_angle")
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    WIND_SPEED_FIELD_NUMBER: _ClassVar[int]
    WIND_ANGLE_FIELD_NUMBER: _ClassVar[int]
    timestamp: int
    wind_speed: int
    wind_angle: int
    def __init__(self, timestamp: _Optional[int] = ..., wind_speed: _Optional[int] = ..., wind_angle: _Optional[int] = ...) -> None: ...



from typing import Optional

from naneos.partector.blueprints._data_structure import NaneosDeviceDataPoint
from naneos.partector_ble.decoder.partector_ble_decoder_blueprint import (
    PartectorBleDecoderBlueprint,
)


class PartectorBleDecoderSize(PartectorBleDecoderBlueprint):
    """
    Decode the std advertisement data from the Partector device.
    """

    # == External used methods =====================================================================
    @classmethod
    def decode(
        cls, data: bytes, data_structure: Optional[NaneosDeviceDataPoint] = None
    ) -> NaneosDeviceDataPoint:
        """
        Decode the auxiliary characteristic data from the Partector device.
        """
        decoded_data = NaneosDeviceDataPoint(
            particle_number_10nm=cls._get_10nm(data),
            particle_number_16nm=cls._get_10nm(data),
            particle_number_26nm=cls._get_26nm(data),
            particle_number_43nm=cls._get_43nm(data),
            particle_number_70nm=cls._get_70nm(data),
            particle_number_114nm=cls._get_114nm(data),
            particle_number_185nm=cls._get_185nm(data),
            particle_number_300nm=cls._get_300nm(data),
        )

        if not data_structure:
            return decoded_data

        # Fill the given data_structure with the decoded data
        for field in NaneosDeviceDataPoint.BLE_SIZE_DIST_FIELD_NAMES:
            setattr(data_structure, field, getattr(decoded_data, field))
        return data_structure

    # == Helpers ===================================================================================
    @staticmethod
    def _get_10nm(data: bytes) -> float:
        """
        Get the 10nm value from the size characteristic data.
        """
        val = float(
            int.from_bytes(bytearray([data[0], data[1], data[2] & 0x0F]), byteorder="little")
        )
        return val

    @staticmethod
    def _get_16nm(data: bytes) -> float:
        """
        Get the 16nm value from the size characteristic data.
        """
        val = float(
            int.from_bytes(bytearray([data[2] & 0xF0, data[3], data[4]]), byteorder="little") >> 4
        )
        return val

    @staticmethod
    def _get_26nm(data: bytes) -> float:
        """
        Get the 26nm value from the size characteristic data.
        """
        val = float(
            int.from_bytes(bytearray([data[5], data[6], data[7] & 0x0F]), byteorder="little")
        )
        return val

    @staticmethod
    def _get_43nm(data: bytes) -> float:
        """
        Get the 43nm value from the size characteristic data.
        """
        val = float(
            int.from_bytes(bytearray([data[7] & 0xF0, data[8], data[9]]), byteorder="little") >> 4
        )
        return val

    @staticmethod
    def _get_70nm(data: bytes) -> float:
        """
        Get the 70nm value from the size characteristic data.
        """
        val = float(
            int.from_bytes(bytearray([data[10], data[11], data[12] & 0x0F]), byteorder="little")
        )
        return val

    @staticmethod
    def _get_114nm(data: bytes) -> float:
        """
        Get the 114nm value from the size characteristic data.
        """
        val = float(
            int.from_bytes(bytearray([data[12] & 0xF0, data[13], data[14]]), byteorder="little")
            >> 4
        )
        return val

    @staticmethod
    def _get_185nm(data: bytes) -> float:
        """
        Get the 185nm value from the size characteristic data.
        """
        val = float(
            int.from_bytes(bytearray([data[15], data[16], data[17] & 0x0F]), byteorder="little")
        )
        return val

    @staticmethod
    def _get_300nm(data: bytes) -> float:
        """
        Get the 300nm value from the size characteristic data.
        """
        val = float(
            int.from_bytes(bytearray([data[17] & 0xF0, data[18], data[19]]), byteorder="little")
            >> 4
        )
        return val



import os
import queue
import signal
import threading
import time

import pandas as pd

from naneos.iotweb.naneos_upload_thread import NaneosUploadThread
from naneos.logger import LEVEL_WARNING, get_naneos_logger
from naneos.partector.blueprints._data_structure import (
    add_to_existing_naneos_data,
    sort_and_clean_naneos_data,
)
from naneos.partector.partector_serial_manager import PartectorSerialManager
from naneos.partector_ble.partector_ble_manager import PartectorBleManager

logger = get_naneos_logger(__name__, LEVEL_WARNING)


class NaneosDeviceManager(threading.Thread):
    """
    NaneosDeviceManager is a class that manages Naneos devices.
    It connects and disconnects automatically.
    """

    def __init__(
        self, use_serial=True, use_ble=True, upload_active=True, gathering_interval_seconds=30
    ) -> None:
        super().__init__(daemon=True)
        self._use_serial = use_serial
        self._use_ble = use_ble
        self._upload_active = upload_active
        self._next_upload_time = time.time() + gathering_interval_seconds
        self.set_gathering_interval_seconds(gathering_interval_seconds)

        self._out_queue: queue.Queue | None = None

        self._stop_event = threading.Event()

        self._manager_serial: PartectorSerialManager | None = None
        self._manager_ble: PartectorBleManager | None = None

        self._data: dict[int, pd.DataFrame] = {}

        self.upload_blocked_devices: list[int | None] = []

    def use_serial_connections(self, use: bool) -> None:
        self._use_serial = use

    def use_ble_connections(self, use: bool) -> None:
        self._use_ble = use

    def get_serial_connection_status(self) -> bool:
        return self._use_serial

    def get_ble_connection_status(self) -> bool:
        return self._use_ble

    def get_upload_status(self) -> bool:
        return self._upload_active

    def set_upload_status(self, active: bool) -> None:
        self._upload_active = active

    def get_gathering_interval_seconds(self) -> int:
        return self._gathering_interval_seconds

    def set_gathering_interval_seconds(self, interval: int) -> None:
        interval = max(10, min(600, interval))
        logger.info(f"Setting gathering interval to {interval} seconds.")
        self._gathering_interval_seconds = interval

        tmp_next_upload_time = time.time() + self._gathering_interval_seconds
        self._next_upload_time = min(self._next_upload_time, tmp_next_upload_time)

    def register_output_queue(self, out_queue: queue.Queue) -> None:
        self._out_queue = out_queue

    def unregister_output_queue(self) -> None:
        self._out_queue = None

    def run(self) -> None:
        self._loop()

        # graceful shutdown in any case
        self._use_serial = False
        self._loop_serial_manager()
        self._use_ble = False
        self._loop_ble_manager()

    def stop(self) -> None:
        self._stop_event.set()

    def get_connected_serial_devices(self) -> list[str]:
        """
        Returns a list of connected serial devices.
        """
        if self._manager_serial is None:
            return []

        return self._manager_serial.get_connected_device_strings()

    def get_connected_ble_devices(self) -> list[str]:
        """
        Returns a list of connected BLE devices.
        """
        if self._manager_ble is None:
            return []

        return self._manager_ble.get_connected_device_strings()

    def get_seconds_until_next_upload(self) -> float:
        """
        Returns the number of seconds until the next upload.
        This is used to determine when to upload data.
        """
        return max(0, self._next_upload_time - time.time())

    def _loop_serial_manager(self) -> None:
        # normal operation
        if (
            isinstance(self._manager_serial, PartectorSerialManager)
            and self._manager_serial.is_alive()
        ):
            self.upload_blocked_devices = self._manager_serial.get_gain_test_activating_devices()
            data_serial = self._manager_serial.get_data()
            self._data = add_to_existing_naneos_data(self._data, data_serial)
        # starting
        if self._manager_serial is None and self._use_serial:
            logger.info("Starting serial manager...")
            self._manager_serial = PartectorSerialManager()
            self._manager_serial.start()
        # stopping
        if isinstance(self._manager_serial, PartectorSerialManager) and not self._use_serial:
            logger.info("Stopping serial manager...")
            self._manager_serial.stop()
            self._manager_serial.join()
            self._manager_serial = None

    def _loop_ble_manager(self) -> None:
        # normal operation
        if isinstance(self._manager_ble, PartectorBleManager) and self._manager_ble.is_alive():
            data_ble = self._manager_ble.get_data()
            self._data = add_to_existing_naneos_data(self._data, data_ble)
        # starting
        if self._manager_ble is None and self._use_ble:
            logger.info("Starting BLE manager...")
            self._manager_ble = PartectorBleManager()
            self._manager_ble.start()
        # stopping
        if isinstance(self._manager_ble, PartectorBleManager) and not self._use_ble:
            logger.info("Stopping BLE manager...")
            self._manager_ble.stop()
            self._manager_ble.join()
            self._manager_ble = None

    def _loop(self) -> None:
        self._next_upload_time = time.time() + self._gathering_interval_seconds

        while not self._stop_event.is_set():
            try:
                time.sleep(1)

                self._loop_serial_manager()
                self._loop_ble_manager()

                # remove entries from _data that is in upload_blocked_devices
                for blocked_sn in self.upload_blocked_devices:
                    if blocked_sn in self._data:
                        del self._data[blocked_sn]

                if time.time() >= self._next_upload_time:
                    self._next_upload_time = time.time() + self._gathering_interval_seconds

                    serial_connected_sns: list[int | None] = []
                    if self._use_serial and self._manager_serial is not None:
                        serial_connected_sns = self._manager_serial.get_connected_serial_numbers()

                    upload_data = sort_and_clean_naneos_data(self._data, serial_connected_sns)
                    self._data = {}

                    if isinstance(self._out_queue, queue.Queue):
                        self._out_queue.put(upload_data)

                    if self._upload_active:
                        uploader = NaneosUploadThread(
                            upload_data,
                            callback=lambda success: logger.info(f"Upload success: {success}"),
                        )
                        uploader.start()
                        uploader.join()

            except Exception as e:
                logger.exception(f"DeviceManager loop exception: {e}")


def minimal_example() -> None:
    manager = NaneosDeviceManager(
        use_serial=True,
        use_ble=True,
        upload_active=False,
        gathering_interval_seconds=10,  # clamped to [10, 600]
    )
    manager.start()

    try:
        while True:
            # Sleep exactly until the next publish window
            remaining = manager.get_seconds_until_next_upload()
            print(f"Next upload in: {remaining:.0f}s")
            time.sleep(remaining + 1)

            print("Serial:", manager.get_connected_serial_devices())
            print("BLE   :", manager.get_connected_ble_devices())
            print()
    except KeyboardInterrupt:
        pass

    manager.stop()
    manager.join()
    print("Stopped.")


def queue_example() -> None:
    import queue

    out_q: queue.Queue = queue.Queue()

    manager = NaneosDeviceManager(
        upload_active=True,
        gathering_interval_seconds=10,
        use_ble=True,
        use_serial=True,
    )
    manager.register_output_queue(out_q)
    manager.start()

    try:
        while True:
            # Wait until a snapshot is ready, then pull all pending ones
            time.sleep(manager.get_seconds_until_next_upload() + 1)

            while not out_q.empty():
                snapshot = out_q.get()
                # snapshot: dict[int, pandas.DataFrame] keyed by device serial
                print(f"Received snapshot for {len(snapshot)} device(s)")
                for serial, df in snapshot.items():
                    print(f"  - {serial}: {len(df)} rows, ldsa: {df['ldsa'].mean():.1f}")
                    # >>> Your processing here (store, analyze, forward, etc.)
                    # print(df.dropna(axis=1, how="all"))
    except KeyboardInterrupt:
        pass

    manager.stop()
    manager.join()


def raise_keyboard_interrupt():
    os.kill(os.getpid(), signal.SIGINT)


def test_naneos_device_manager():
    timer = threading.Timer(300, raise_keyboard_interrupt)  # trigger keyboard interrupt after 20s
    timer.start()

    manager = NaneosDeviceManager()
    manager.start()

    try:
        while True:
            time.sleep(1)
            print(f"Seconds until next upload: {manager.get_seconds_until_next_upload():.0f}")
            print(manager.get_connected_serial_devices())
            print(manager.get_connected_ble_devices())
            print()
    except KeyboardInterrupt:
        manager.stop()
        manager.join()
        print("NaneosDeviceManager stopped.")

    timer.cancel()


def ble_connect_example() -> None:
    manager = NaneosDeviceManager(
        use_serial=False,
        use_ble=True,
        upload_active=False,
        gathering_interval_seconds=10,  # clamped to [10, 600]
    )
    manager.start()

    try:
        while True:
            # check if 8617 is connected
            devices = manager.get_connected_ble_devices()
            if "SN8617" in devices:
                print("8617 is connected")
                manager.stop()
                manager.join()
                print("Stopped.")

                # restart
                manager = NaneosDeviceManager(
                    use_serial=False,
                    use_ble=True,
                    upload_active=False,
                    gathering_interval_seconds=10,  # clamped to [10, 600]
                )
                manager.start()
                print("Restarted.")

    except KeyboardInterrupt:
        pass

    manager.stop()
    manager.join()
    print("Stopped.")


if __name__ == "__main__":
    # minimal_example()
    queue_example()
    # test_naneos_device_manager()
    # ble_connect_example()

    # df = pd.read_pickle("partector_data_sn24.pkl")
    # print(df)



from __future__ import annotations

import asyncio
import time
from typing import Optional

from bleak import BleakClient
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.exc import BleakDeviceNotFoundError

from naneos.logger import LEVEL_WARNING, get_naneos_logger
from naneos.partector.blueprints._data_structure import NaneosDeviceDataPoint
from naneos.partector_ble.decoder.partectod_ble_decoder_aux_error import PartectorBleDecoderAuxError
from naneos.partector_ble.decoder.partector_ble_decoder_aux import PartectorBleDecoderAux
from naneos.partector_ble.decoder.partector_ble_decoder_size import PartectorBleDecoderSize
from naneos.partector_ble.decoder.partector_ble_decoder_std import PartectorBleDecoderStd

logger = get_naneos_logger(__name__, LEVEL_WARNING)


class PartectorBleConnection:
    SERVICE_UUID = "0bd51666-e7cb-469b-8e4d-2742f1ba77cc"
    CHAR_UUIDS = {
        "std": "e7add780-b042-4876-aae1-112855353cc1",
        "aux": "e7add781-b042-4876-aae1-112855353cc1",
        "write": "e7add782-b042-4876-aae1-112855353cc1",
        "read": "e7add783-b042-4876-aae1-112855353cc1",
        "size_dist": "e7add784-b042-4876-aae1-112855353cc1",
    }

    # static methods ###############################################################################
    @staticmethod
    def create_connection_queue() -> asyncio.Queue[NaneosDeviceDataPoint]:
        """Create a queue for the connection data."""
        # Increased maxsize to 500 to handle bursts from multiple devices
        # Prevents message loss on Raspberry Pi with many concurrent connections
        queue_connection: asyncio.Queue[NaneosDeviceDataPoint] = asyncio.Queue(maxsize=500)

        return queue_connection

    # == Lifecycle and Context Management ==========================================================
    def __init__(
        self,
        device: BLEDevice,
        loop: asyncio.AbstractEventLoop,
        serial_number: int,
        queue: asyncio.Queue[NaneosDeviceDataPoint],
    ) -> None:
        """
        Initializes the BLE connection with the given device, event loop, and queue.

        Args:
            device (BLEDevice): The BLE device to connect to.
            loop (asyncio.AbstractEventLoop): The event loop to run the connection in.
            serial_number (int): The serial number of the device.
        """
        self.SERIAL_NUMBER = serial_number
        self._device_type = NaneosDeviceDataPoint.DEV_TYPE_P2  # Thats the deafault value
        self._data = NaneosDeviceDataPoint()
        self._next_ts = 0.0
        self._last_aux_data_ts = time.time()
        self._queue = queue

        # Decode queue to decouple decoding from BLE callbacks
        # This prevents blocking the event loop when decoding heavy data
        self._decode_queue: asyncio.Queue = asyncio.Queue(maxsize=200)

        self._device = device
        self._loop = loop
        self._task: asyncio.Task | None = None
        self._stop_event = asyncio.Event()
        self._stop_event.set()  # stopped by default
        self._client = BleakClient(device, self._disconnect_callback, timeout=10)

    async def __aenter__(self) -> PartectorBleConnection:
        self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.stop()

    # == Public Methods ============================================================================
    def start(self) -> None:
        """Starts the scanner."""
        if not self._stop_event.is_set():
            logger.warning("SN{self._serial_number}: start() called while already running")
            return
        self._stop_event.clear()
        self._task = self._loop.create_task(self._run())

    async def stop(self) -> None:
        """Stops the scanner."""
        self._stop_event.set()
        if self._task and not self._task.done():
            await self._task
        logger.info(f"SN{self.SERIAL_NUMBER}: PartectorBleConnection stopped")

    async def _run(self) -> None:
        waiting_seconds = 0

        try:
            self._next_ts = int(time.time()) + 1.0

            # Create decode task to run in parallel
            # This prevents decoding from blocking the event loop
            self._loop.create_task(self._decode_routine())

            while not self._stop_event.is_set():
                try:
                    if self._last_aux_data_ts + 60 < time.time():
                        logger.info(
                            f"SN{self.SERIAL_NUMBER}: No aux data received for 60 seconds, disconnecting to reset."
                        )
                        await self._disconnect_gracefully()
                        self._last_aux_data_ts = time.time()
                        waiting_seconds = 5  # wait 5 seconds before reconnecting

                    waiting_seconds = max(0, waiting_seconds - 1)
                    wait = self._next_ts - time.time()
                    if wait > 0:
                        await asyncio.sleep(wait)
                        self._next_ts += 1.0
                    else:
                        if self._client.is_connected:
                            logger.info(f"SN{self.SERIAL_NUMBER}: Waiting time negative: {wait}")
                        self._next_ts = int(time.time()) + 1.0

                    if self._client.is_connected:
                        if (
                            self._device_type == NaneosDeviceDataPoint.DEV_TYPE_P2PRO
                            and self._data.particle_number_10nm is None
                            and self._data.particle_number_16nm is None
                            and self._data.particle_number_26nm is None
                            and self._data.particle_number_43nm is None
                            and self._data.particle_number_70nm is None
                            and self._data.particle_number_114nm is None
                            and self._data.particle_number_185nm is None
                            and self._data.particle_number_300nm is None
                        ):
                            self._data.particle_number_concentration = None
                            self._data.average_particle_diameter = None

                        self._queue.put_nowait(self._data)
                        self._data = NaneosDeviceDataPoint(
                            device_type=self._device_type,
                            serial_number=self.SERIAL_NUMBER,
                            connection_type=NaneosDeviceDataPoint.CONN_TYPE_CONNECTED,
                            # TODO: add firware version from device here
                        )
                        continue

                    if waiting_seconds == 0:
                        await self._client.connect(timeout=5)  # 5 seconds for windows...
                        if self._client.is_connected:
                            await self._client.start_notify(
                                self.CHAR_UUIDS["std"], self._callback_std
                            )
                            await self._client.start_notify(
                                self.CHAR_UUIDS["aux"], self._callback_aux
                            )
                            await self._client.start_notify(
                                self.CHAR_UUIDS["size_dist"], self._callback_size_dist
                            )
                        logger.info(f"SN{self.SERIAL_NUMBER}: Connected to {self._device.address}")

                    self._next_ts = int(time.time()) + 1.0
                except asyncio.TimeoutError:
                    logger.info(f"SN{self.SERIAL_NUMBER}: Connection timeout.")
                    waiting_seconds = 30
                    await asyncio.sleep(0.5)
                except BleakDeviceNotFoundError:
                    logger.info(f"SN{self.SERIAL_NUMBER}: Device not found or probably old BLE.")
                    waiting_seconds = 30
                    await asyncio.sleep(0.5)
                except Exception as e:
                    # if exception contains "not found" increase waiting time to 30 seconds and do not spam
                    if "not found" in str(e).lower():
                        logger.info(
                            f"SN{self.SERIAL_NUMBER}: Device not found or probably old BLE: {e}"
                        )
                        waiting_seconds = 30
                    else:
                        logger.warning(f"SN{self.SERIAL_NUMBER}: Unknown exception: {e}")

                    await asyncio.sleep(0.5)
        except asyncio.CancelledError:
            logger.warning(f"SN{self.SERIAL_NUMBER}: _run task cancelled.")
        except Exception as e:
            logger.exception(f"SN{self.SERIAL_NUMBER}: _run task failed: {e}")
        finally:
            await self._disconnect_gracefully()

    async def _decode_routine(self) -> None:
        """Asynchronously decodes BLE data from the decode queue.

        This runs in parallel with the main connection loop, preventing
        decoding from blocking the event loop when handling multiple connections.
        """
        while not self._stop_event.is_set():
            try:
                # Non-blocking check with timeout to allow graceful shutdown
                try:
                    char_type, data = await asyncio.wait_for(self._decode_queue.get(), timeout=0.5)
                except asyncio.TimeoutError:
                    continue

                # Update timestamp for all decodings
                self._data.unix_timestamp = int(time.time() * 1000)

                # Decode based on characteristic type
                if char_type == "std":
                    self._data = PartectorBleDecoderStd.decode(data, data_structure=self._data)
                    logger.debug(f"SN{self.SERIAL_NUMBER}: Decoded std: {data.hex()}")

                elif char_type == "aux":
                    # Check for aux error data
                    if len(data) >= 2 and data[0] == 255 and data[1] == 255:
                        self._data = PartectorBleDecoderAuxError.decode(
                            data, data_structure=self._data
                        )
                    else:
                        self._data = PartectorBleDecoderAux.decode(data, data_structure=self._data)
                    logger.debug(f"SN{self.SERIAL_NUMBER}: Decoded aux: {data.hex()}")

                elif char_type == "size_dist":
                    self._device_type = NaneosDeviceDataPoint.DEV_TYPE_P2PRO
                    self._data = PartectorBleDecoderSize.decode(data, data_structure=self._data)
                    logger.debug(f"SN{self.SERIAL_NUMBER}: Decoded size_dist: {data.hex()}")

            except Exception as e:
                logger.warning(f"SN{self.SERIAL_NUMBER}: Error in decode routine: {e}")

    async def _disconnect_gracefully(self) -> None:
        if not self._client.is_connected:
            return

        try:
            await asyncio.wait_for(self._client.stop_notify(self.CHAR_UUIDS["std"]), timeout=1)
            await asyncio.sleep(0.5)  # wait for windows to free resources
            await asyncio.wait_for(self._client.stop_notify(self.CHAR_UUIDS["aux"]), timeout=1)
            await asyncio.sleep(0.5)  # wait for windows to free resources
            await asyncio.wait_for(
                self._client.stop_notify(self.CHAR_UUIDS["size_dist"]), timeout=1
            )
            await asyncio.sleep(0.5)  # wait for windows to free resources
        except Exception as e:
            logger.debug(f"SN{self.SERIAL_NUMBER}: Failed to stop notify: {e}")

        try:
            await asyncio.wait_for(self._client.disconnect(), timeout=1)
            await asyncio.sleep(0.5)  # wait for windows to free resources
        except Exception as e:
            logger.debug(f"SN{self.SERIAL_NUMBER}: Failed to disconnect: {e}")

    def _disconnect_callback(self, client: BleakClient) -> None:
        """Callback on disconnect."""
        logger.debug(f"SN{self.SERIAL_NUMBER}: Disconnect callback called")

    def _callback_std(self, characteristic: BleakGATTCharacteristic, data: bytearray) -> None:
        """Callback on data received (std characteristic).

        Non-blocking: puts data in decode queue instead of decoding directly.
        Actual decoding happens asynchronously in _decode_routine().
        """
        try:
            self._decode_queue.put_nowait(("std", bytes(data)))
        except asyncio.QueueFull:
            logger.warning(f"SN{self.SERIAL_NUMBER}: Decode queue full, dropping std data")

    def _callback_aux(self, characteristic: BleakGATTCharacteristic, data: bytearray) -> None:
        """Callback on data received (aux characteristic).

        Non-blocking: puts data in decode queue instead of decoding directly.
        Actual decoding happens asynchronously in _decode_routine().
        """
        self._last_aux_data_ts = time.time()
        try:
            self._decode_queue.put_nowait(("aux", bytes(data)))
        except asyncio.QueueFull:
            logger.warning(f"SN{self.SERIAL_NUMBER}: Decode queue full, dropping aux data")

    def _callback_size_dist(self, characteristic: BleakGATTCharacteristic, data: bytearray) -> None:
        """Callback on data received (size_dist characteristic).

        Non-blocking: puts data in decode queue instead of decoding directly.
        Actual decoding happens asynchronously in _decode_routine().
        """
        try:
            self._decode_queue.put_nowait(("size_dist", bytes(data)))
        except asyncio.QueueFull:
            logger.warning(f"SN{self.SERIAL_NUMBER}: Decode queue full, dropping size_dist data")


async def main():
    from naneos.partector_ble.partector_ble_scanner import PartectorBleScanner

    SNS = {8112, 8617}
    conn_list = []  # serial number to connection mapping

    loop = asyncio.get_event_loop()
    queue_scanner = PartectorBleScanner.create_scanner_queue()
    queue_connection = PartectorBleConnection.create_connection_queue()

    async with PartectorBleScanner(loop=loop, queue=queue_scanner):
        await asyncio.sleep(5)

    device_dict = await _map_sn_to_device(queue_scanner)
    if not device_dict:
        return

    device_dict = {k: v for k, v in device_dict.items() if k in SNS}

    # start connections for all devices
    for serial_number, device in device_dict.items():
        conn_list.append(
            PartectorBleConnection(
                device=device, loop=loop, serial_number=serial_number, queue=queue_connection
            )
        )
        conn_list[-1].start()

    await asyncio.sleep(10)

    # stop connections for all devices
    for conn in conn_list:
        await conn.stop()

    # print the data from the queue
    while not queue_connection.empty():
        data = await queue_connection.get()
        print(data)


async def _map_sn_to_device(
    queue: asyncio.Queue[tuple[BLEDevice, NaneosDeviceDataPoint]],
) -> Optional[dict[int, BLEDevice]]:
    device_dict = {}
    while not queue.empty():
        device, data = await queue.get()
        if data.serial_number:
            device_dict[data.serial_number] = device

    if not device_dict:
        logger.info("No devices found.")
        return None

    return device_dict


async def main_x(x):
    for _ in range(x):
        await main()


if __name__ == "__main__":
    asyncio.run(main_x(3))



import time

from naneos.manager import NaneosDeviceManager


def minimal_example() -> None:
    manager = NaneosDeviceManager(
        use_serial=True,
        use_ble=True,
        upload_active=False,
        gathering_interval_seconds=10,  # clamped to [10, 600]
    )
    manager.start()

    try:
        while True:
            # Sleep exactly until the next publish window
            remaining = manager.get_seconds_until_next_upload()
            print(f"Next upload in: {remaining:.0f}s")
            time.sleep(remaining + 1)

            print("Serial:", manager.get_connected_serial_devices())
            print("BLE   :", manager.get_connected_ble_devices())
            print()
    except KeyboardInterrupt:
        pass

    manager.stop()
    manager.join()
    print("Stopped.")


def queue_example() -> None:
    import queue

    out_q: queue.Queue = queue.Queue()

    manager = NaneosDeviceManager(
        upload_active=True,
        gathering_interval_seconds=10,
        use_ble=True,
        use_serial=True,
    )
    manager.register_output_queue(out_q)
    manager.start()

    try:
        while True:
            time.sleep(manager.get_seconds_until_next_upload() + 1)

            while not out_q.empty():
                snapshot = out_q.get()
                print(f"Received snapshot for {len(snapshot)} device(s)")
                for serial, df in snapshot.items():
                    print(f"  - {serial}: {len(df)} rows, ldsa: {df['ldsa'].mean():.1f}")
    except KeyboardInterrupt:
        pass

    manager.stop()
    manager.join()


if __name__ == "__main__":
    queue_example()



import asyncio
import warnings

from bleak.backends.device import BLEDevice

from naneos.partector.blueprints._data_structure import NaneosDeviceDataPoint
from naneos.partector_ble.partector_ble_scanner import PartectorBleScanner


async def check_queue(
    queue: asyncio.Queue[tuple[BLEDevice, NaneosDeviceDataPoint]],
) -> None:
    if queue.empty():
        warnings.warn("No BLE devices found.", UserWarning)

    # check if there is data in the queue
    while not queue.empty():
        device, data = await queue.get()
        assert device is not None
        assert data is not None
        assert data.serial_number is not None


async def async_test_scanner(with_context_manager: bool) -> None:
    """Helper function to test the scanner."""
    loop = asyncio.get_event_loop()
    queue_scanner = PartectorBleScanner.create_scanner_queue()

    if with_context_manager:
        async with PartectorBleScanner(loop=loop, queue=queue_scanner) as scanner:
            await asyncio.sleep(3)
    else:
        scanner = PartectorBleScanner(loop=loop, queue=queue_scanner)
        scanner.start()
        try:
            await asyncio.sleep(3)  # Asynchronous sleep to allow the loop to run
        finally:
            await scanner.stop()

    await check_queue(queue_scanner)


def test_scanner() -> None:
    """Test the scanner functionality."""
    asyncio.run(async_test_scanner(with_context_manager=False))


def test_scanner_with_context_manager() -> None:
    """Test the scanner functionality with context manager."""
    asyncio.run(async_test_scanner(with_context_manager=True))



import time
from typing import Optional

import pandas as pd

from naneos.partector.blueprints._data_structure import (
    PARTECTOR2_DATA_STRUCTURE_V320,
    PARTECTOR2_GAIN_TEST_ADDITIONAL_DATA_STRUCTURE,
    PARTECTOR2_OUTPUT_PULSE_DIAGNOSTIC_ADDITIONAL_DATA_STRUCTURE,
    PARTECTOR2_PRO_DATA_STRUCTURE_V311,
    PARTECTOR2_PRO_DATA_STRUCTURE_V336,
    NaneosDeviceDataPoint,
)
from naneos.partector.blueprints._partector_blueprint import PartectorBluePrint


class Partector2Pro(PartectorBluePrint):
    def __init__(
        self,
        serial_number: Optional[int] = None,
        port: Optional[str] = None,
        verb_freq: int = 6,
        hw_version: str = "P2pro",
        gain_test_active: bool = True,
        output_pulse_diagnostics: bool = True,
    ) -> None:
        self._GAIN_TEST_ACTIVE = gain_test_active
        self._OUTPUT_PULSE_DIAGNOSTICS = output_pulse_diagnostics
        super().__init__(serial_number, port, verb_freq, hw_version)

    def _init_serial_data_structure(self) -> None:
        """This gets passed here and is set in the set_verbose_freq method."""
        self.device_type = NaneosDeviceDataPoint.DEV_TYPE_P2PRO

    def _set_verbose_freq(self, freq: int) -> None:
        if freq == 0:
            self._write_line("X0000!")
        elif freq in [1, 2, 3]:  # std p2 mode
            if self._fw >= 311:
                self._data_structure = PARTECTOR2_DATA_STRUCTURE_V320

                self._write_line("M0000!")  # deactivates size dist mode
                self._write_line("A0002!")  # activates antispikes

                if self._OUTPUT_PULSE_DIAGNOSTICS:
                    self._write_line("opd01!")
                    self._data_structure.update(
                        PARTECTOR2_OUTPUT_PULSE_DIAGNOSTIC_ADDITIONAL_DATA_STRUCTURE
                    )
                else:
                    self._write_line("opd00!")

                if self._GAIN_TEST_ACTIVE:
                    waiting_time = max(10, (self._integration_time + 5))
                    self._wait_with_data_output_until = time.time() + waiting_time
                    self._write_line("h2001!")  # activates harmonics output
                    self._write_line("e1100!")  # strength of gain test signal
                    self._data_structure.update(PARTECTOR2_GAIN_TEST_ADDITIONAL_DATA_STRUCTURE)
                else:
                    self._write_line("h2000!")  # deactivates harmonics output
                    self._write_line("e0000!")  # deactivates gain test signal

                self._write_line(f"X000{freq}!")  # set verbose freq

            else:
                raise RuntimeError("Firmware too old for P2 pro mode. Minimum FW is 311.")
        elif freq == 6:  # p2 pro mode
            if self._fw >= 336:
                self._data_structure = PARTECTOR2_PRO_DATA_STRUCTURE_V336
            else:
                self._data_structure = PARTECTOR2_PRO_DATA_STRUCTURE_V311

            self._write_line("X0006!")  # activates verbose mode
            self._write_line("M0004!")  # activates size dist mode
            self._write_line("A0002!")  # activates the antispikes

            if self._OUTPUT_PULSE_DIAGNOSTICS:
                self._write_line("opd01!")
                self._data_structure.update(
                    PARTECTOR2_OUTPUT_PULSE_DIAGNOSTIC_ADDITIONAL_DATA_STRUCTURE
                )
            else:
                self._write_line("opd00!")

            if self._GAIN_TEST_ACTIVE:
                waiting_time = max(10, (self._integration_time + 5))
                self._wait_with_data_output_until = time.time() + waiting_time
                self._write_line("h2001!")  # activates harmonics output
                self._write_line("e1100!")  # strength of gain test signal
                self._data_structure.update(PARTECTOR2_GAIN_TEST_ADDITIONAL_DATA_STRUCTURE)
            else:
                self._write_line("h2000!")  # deactivates harmonics output
                self._write_line("e0000!")  # deactivates gain test signal


if __name__ == "__main__":
    from naneos.partector.scanPartector import scan_for_serial_partectors

    partectors = scan_for_serial_partectors()
    assert partectors["P2pro"], "No Partector found!"
    serial_number, port = next(iter(partectors["P2pro"].items()))

    data: dict[int, pd.DataFrame] = {}
    p2_pro = Partector2Pro(port=port, verb_freq=6, gain_test_active=True)

    for _ in range(5):
        time.sleep(3)
        data_points = p2_pro.get_data()
        for point in data_points:
            data = NaneosDeviceDataPoint.add_data_point_to_dict(data, point)

        df = next(iter(data.values()), pd.DataFrame())
        if not df.empty:
            print(f"Sn: {p2_pro._sn}, Port: {p2_pro._port}")
            print(df.dropna(axis=1, how="all"))
            # print(df["diffusion_current_delay_on"])
            break

        print("No data received yet...")

    p2_pro.close(blocking=True)



# import datetime as dt
# import os

# import pandas as pd

# from naneos.iotweb import download_from_iotweb
# from naneos.iotweb.naneos_upload_thread import NaneosUploadThread


# def test_download_8134() -> None:
#     token: str | None = os.getenv("IOT_GUEST_TOKEN", None)
#     if token is None:
#         raise ValueError("No token found in your environment")

#     # use local timezone for start
#     start = dt.datetime(2025, 4, 1)
#     stop = dt.datetime(2025, 4, 7)
#     serial_number = "8134"
#     name = "iot_guest"

#     df: pd.DataFrame = download_from_iotweb(name, serial_number, start, stop, token)
#     assert len(df) == 206545


# callback_upload_success = {"result": False}


# def callback_upload(ret: bool) -> None:
#     callback_upload_success["result"] = ret


# def test_upload() -> None:
#     data_dir = os.path.join(os.path.dirname(__file__), "data")

#     df_p2 = pd.read_pickle(os.path.join(data_dir, "p2_test_data.pkl"))
#     assert not df_p2.empty, "DataFrame is empty or not loaded correctly"

#     df_p2_pro = pd.read_pickle(os.path.join(data_dir, "p2_pro_test_data.pkl"))
#     assert not df_p2_pro.empty, "DataFrame is empty or not loaded correctly"

#     data = [
#         (int(777), str("P2pro"), df_p2_pro),
#         (int(666), str("P2"), df_p2),
#     ]

#     thread = NaneosUploadThread(data, callback_upload)

#     thread.start()
#     thread.join()

#     assert callback_upload_success["result"], "Upload failed, callback returned False"


# if __name__ == "__main__":
#     # test_download_8134()
#     test_upload()



import time

import serial

from naneos.partector.scanPartector import scan_for_serial_partectors


### Utility to send commands from a commands file to a device via serial connection.
def send_commands_to_device(serial_number: int, commands_path: str) -> None:
    scan_result = scan_for_serial_partectors()
    scan_result = {k: v for x in scan_result.values() for (k, v) in x.items()}  #

    if serial_number not in scan_result:
        raise ValueError(f"Device with serial number {serial_number} not found.")
    port = scan_result[serial_number]

    with serial.Serial(port, baudrate=9600, timeout=1) as ser:
        ser.flush()
        ser.write(b"!")  # clear device buffer
        time.sleep(1)  # wait a bit

        with open(commands_path, "r") as f:
            for line in f:
                if line.startswith("2"):
                    ser.write(line.encode())
                    time.sleep(0.1)  # slight delay between commands


if __name__ == "__main__":
    send_commands_to_device(8617, "/Users/huegi/Downloads/commandsSN8617.txt")



import nox

# Stelle sicher, dass uv als Backend verwendet wird
nox.options.default_venv_backend = "uv"


@nox.session(python=["3.10", "3.11", "3.12", "3.13"])
def tests(session):
    session.install(".[test]")
    session.run("pytest")
    # session.run("coverage", "run", "-m", "pytest")
    # session.run("coverage", "report", "-m")



import signal
import time

from naneos.manager import NaneosDeviceManager

running = True  # global flag to control the main loop


def handle_signal(signum, frame):
    global running
    running = False


# register signal handlers for SIGTERM and SIGINT
signal.signal(signal.SIGTERM, handle_signal)
signal.signal(signal.SIGINT, handle_signal)


def rp_service_main() -> None:
    manager = NaneosDeviceManager(
        use_serial=True, use_ble=True, upload_active=True, gathering_interval_seconds=30
    )
    manager.start()

    try:
        while running:
            remaining = manager.get_seconds_until_next_upload()

            slept = 0
            while running and slept < remaining + 1:
                time.sleep(1)
                slept += 1

            if not running:
                break

    finally:
        manager.stop()
        manager.join()


if __name__ == "__main__":
    rp_service_main()
